/*
Copyright 2020-2023 CAcert Inc.
SPDX-License-Identifier: Apache-2.0

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    https://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
*/

package handlers

import (
	"fmt"
	"html/template"
	"net/http"
	"net/url"

	"github.com/nicksnyder/go-i18n/v2/i18n"
	log "github.com/sirupsen/logrus"

	"code.cacert.org/cacert/oidc-demo-app/ui"

	"code.cacert.org/cacert/oidc-demo-app/internal/services"
)

type IndexHandler struct {
	bundle         *i18n.Bundle
	indexTemplate  *template.Template
	logger         *log.Logger
	logoutURL      string
	messageCatalog *services.MessageCatalog
	publicURL      string
	tokenInfo      *services.TokenInfoService
}

func (h *IndexHandler) GetBundle() *i18n.Bundle {
	return h.bundle
}

func (h *IndexHandler) GetCatalog() *services.MessageCatalog {
	return h.messageCatalog
}

func (h *IndexHandler) ServeHTTP(w http.ResponseWriter, r *http.Request) {
	if r.Method != http.MethodGet {
		http.Error(w, http.StatusText(http.StatusMethodNotAllowed), http.StatusMethodNotAllowed)

		return
	}

	if r.URL.Path != "/" {
		http.NotFound(w, r)

		return
	}

	localizer := GetLocalizer(h, r)

	logoutURL, err := url.Parse(h.logoutURL)
	if err != nil {
		http.Error(w, err.Error(), http.StatusInternalServerError)

		return
	}

	session, err := GetSession(r)
	if err != nil {
		http.Error(w, err.Error(), http.StatusInternalServerError)

		return
	}

	tokenInfo, err := h.tokenInfo.GetTokenInfo(session)
	if err != nil {
		h.logger.WithError(err).Error("failed to get token info for request")

		http.Error(w, err.Error(), http.StatusInternalServerError)

		return
	}

	if !tokenInfo.Expires.IsZero() {
		h.logger.WithField("expires", tokenInfo.Expires).Info("id token expires at")
	}

	w.Header().Add("Content-Type", "text/html")

	msgLookup := h.messageCatalog.LookupMessage

	data := BaseTemplateData(h, localizer)

	data["Title"] = msgLookup("IndexTitle", nil, localizer)
	data["Greeting"] = msgLookup("GreetingAnonymous", nil, localizer)
	data["LoginLabel"] = msgLookup("LoginLabel", nil, localizer)
	data["IntroductionText"] = msgLookup("IndexIntroductionText", nil, localizer)
	data["IsAuthenticated"] = false

	if tokenInfo.IDToken != "" {
		logoutURL.RawQuery = url.Values{
			"id_token_hint":            []string{tokenInfo.IDToken},
			"post_logout_redirect_uri": []string{fmt.Sprintf("%s/after-logout", h.publicURL)},
		}.Encode()

		data["Greeting"] = msgLookup(
			"GreetingAuthenticated", map[string]interface{}{"Name": tokenInfo.Name}, localizer,
		)
		data["LogoutLabel"] = msgLookup("LogoutLabel", nil, localizer)
		data["LogoutURL"] = logoutURL.String()
		data["IsAuthenticated"] = true
		data["AuthenticatedAs"] = msgLookup("AuthenticatedAs", map[string]interface{}{
			"Name":  tokenInfo.Name,
			"Email": tokenInfo.Email,
		}, localizer)
	}

	err = h.indexTemplate.Lookup("base").Execute(w, data)
	if err != nil {
		http.Error(w, err.Error(), http.StatusInternalServerError)

		return
	}
}

func NewIndexHandler(
	logger *log.Logger,
	bundle *i18n.Bundle,
	catalog *services.MessageCatalog,
	oidcInfo *services.OIDCInformation,
	publicURL string,
	tokenInfoService *services.TokenInfoService,
) (*IndexHandler, error) {
	indexTemplate, err := template.ParseFS(
		ui.Templates,
		"templates/base.gohtml", "templates/index.gohtml")
	if err != nil {
		return nil, fmt.Errorf("could not parse templates: %w", err)
	}

	return &IndexHandler{
		logger:         logger,
		bundle:         bundle,
		indexTemplate:  indexTemplate,
		logoutURL:      oidcInfo.OIDCConfiguration.EndSessionEndpoint,
		tokenInfo:      tokenInfoService,
		messageCatalog: catalog,
		publicURL:      publicURL,
	}, nil
}
