/*
Copyright 2020-2023 CAcert Inc.
SPDX-License-Identifier: Apache-2.0

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    https://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
*/

package handlers

import (
	"fmt"
	"html/template"
	"net/http"
	"net/url"

	"github.com/nicksnyder/go-i18n/v2/i18n"
	log "github.com/sirupsen/logrus"

	"code.cacert.org/cacert/oidc-demo-app/internal/services"
	"code.cacert.org/cacert/oidc-demo-app/ui"
)

type ProtectedResource struct {
	bundle            *i18n.Bundle
	logger            *log.Logger
	protectedTemplate *template.Template
	logoutURL         string
	tokenInfo         *services.TokenInfoService
	messageCatalog    *services.MessageCatalog
	publicURL         string
}

func (h *ProtectedResource) GetBundle() *i18n.Bundle {
	return h.bundle
}

func (h *ProtectedResource) GetCatalog() *services.MessageCatalog {
	return h.messageCatalog
}

func (h *ProtectedResource) ServeHTTP(w http.ResponseWriter, r *http.Request) {
	if r.Method != http.MethodGet {
		http.Error(w, http.StatusText(http.StatusMethodNotAllowed), http.StatusMethodNotAllowed)

		return
	}

	if r.URL.Path != "/protected" {
		http.NotFound(w, r)

		return
	}

	localizer := GetLocalizer(h, r)

	logoutURL, err := url.Parse(h.logoutURL)
	if err != nil {
		http.Error(w, err.Error(), http.StatusInternalServerError)

		return
	}

	session, err := GetSession(r)
	if err != nil {
		http.Error(w, err.Error(), http.StatusInternalServerError)

		return
	}

	tokenInfo, err := h.tokenInfo.GetTokenInfo(session)
	if err != nil {
		h.logger.WithError(err).Error("failed to get token info for request")

		http.Error(w, err.Error(), http.StatusInternalServerError)

		return
	}

	if !tokenInfo.Expires.IsZero() {
		h.logger.WithField("expires", tokenInfo.Expires).Info("id token expires at")
	}

	if tokenInfo.IDToken == "" {
		http.Error(w, http.StatusText(http.StatusForbidden), http.StatusForbidden)

		return
	}

	w.Header().Add("Content-Type", "text/html")

	msgLookup := h.messageCatalog.LookupMessage

	logoutURL.RawQuery = url.Values{
		"id_token_hint":            []string{tokenInfo.IDToken},
		"post_logout_redirect_uri": []string{fmt.Sprintf("%s/after-logout", h.publicURL)},
	}.Encode()

	data := BaseTemplateData(h, localizer)

	data["Title"] = msgLookup("IndexTitle", nil, localizer)
	data["Greeting"] = msgLookup(
		"GreetingAuthenticated", map[string]interface{}{"Name": tokenInfo.Name}, localizer,
	)
	data["LoginLabel"] = msgLookup("LoginLabel", nil, localizer)
	data["IntroductionText"] = msgLookup("ProtectedIntroductionText", nil, localizer)
	data["IsAuthenticated"] = true
	data["LogoutLabel"] = msgLookup("LogoutLabel", nil, localizer)
	data["LogoutURL"] = logoutURL.String()
	data["AuthenticatedAs"] = msgLookup("AuthenticatedAs", map[string]interface{}{
		"Name":  tokenInfo.Name,
		"Email": tokenInfo.Email,
	}, localizer)

	err = h.protectedTemplate.Lookup("base").Execute(w, data)
	if err != nil {
		http.Error(w, err.Error(), http.StatusInternalServerError)

		return
	}
}

func NewProtectedResourceHandler(
	logger *log.Logger,
	bundle *i18n.Bundle,
	catalog *services.MessageCatalog,
	oidcInfo *services.OIDCInformation,
	publicURL string, tokenInfoService *services.TokenInfoService,
) (*ProtectedResource, error) {
	protectedTemplate, err := template.ParseFS(
		ui.Templates,
		"templates/base.gohtml", "templates/protected.gohtml")
	if err != nil {
		return nil, fmt.Errorf("could not parse templates: %w", err)
	}

	return &ProtectedResource{
		logger:            logger,
		bundle:            bundle,
		protectedTemplate: protectedTemplate,
		logoutURL:         oidcInfo.OIDCConfiguration.EndSessionEndpoint,
		tokenInfo:         tokenInfoService,
		messageCatalog:    catalog,
		publicURL:         publicURL,
	}, nil
}
