/*
Copyright 2020-2023 CAcert Inc.
SPDX-License-Identifier: Apache-2.0

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    https://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
*/

package handlers

import (
	"context"
	"errors"
	"net/http"
	"os"
	"os/signal"
	"sync/atomic"
	"time"

	"github.com/knadh/koanf"
	"github.com/sirupsen/logrus"
)

func StartApplication(
	ctx context.Context,
	logger *logrus.Logger,
	server *http.Server,
	publicURL string,
	config *koanf.Koanf,
) {
	done := make(chan bool)
	quit := make(chan os.Signal, 1)
	signal.Notify(quit, os.Interrupt)

	go func() {
		<-quit
		logger.Infoln("Server is shutting down...")
		atomic.StoreInt32(&Healthy, 0)

		const shutdownWaitTime = 30 * time.Second

		ctx, cancel := context.WithTimeout(ctx, shutdownWaitTime)

		defer cancel()

		server.SetKeepAlivesEnabled(false)

		if err := server.Shutdown(ctx); err != nil {
			logger.WithError(err).Fatal("Could not gracefully shutdown the server")
		}

		close(done)
	}()

	logger.WithField("public_url", publicURL).Info("Server is ready to handle requests")
	atomic.StoreInt32(&Healthy, 1)

	if err := server.ListenAndServeTLS(
		config.String("server.certificate"), config.String("server.key"),
	); err != nil && !errors.Is(err, http.ErrServerClosed) {
		logger.WithError(err).WithField(
			"server_address",
			server.Addr,
		).Fatal("Could not listen on requested address")
	}

	<-done
	logger.Infoln("Server stopped")
}
