/*
Copyright 2020-2023 CAcert Inc.
SPDX-License-Identifier: Apache-2.0

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    https://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
*/

package services

import (
	"fmt"
	"os"
	"strings"

	"github.com/knadh/koanf"
	"github.com/knadh/koanf/parsers/toml"
	"github.com/knadh/koanf/providers/confmap"
	"github.com/knadh/koanf/providers/env"
	"github.com/knadh/koanf/providers/file"
	"github.com/knadh/koanf/providers/posflag"
	"github.com/sirupsen/logrus"
	"github.com/spf13/pflag"
)

const (
	defaultServerPort = 4000
	defaultFile       = "resource_app.toml"
)

var DefaultConfiguration = map[string]interface{}{
	"server.bind_address": "",
	"server.name":         "app.cacert.localhost",
	"server.port":         defaultServerPort,
	"server.key":          "certs/app.cacert.localhost.key",
	"server.certificate":  "certs/app.cacert.localhost.crt.pem",
	"oidc.server":         "https://auth.cacert.localhost:4444/",
	"session.path":        "sessions/app",
	"i18n.languages":      []string{"en", "de"},
	"log.level":           "info",
	"log.json":            true,
}

func ConfigureApplication(
	logger *logrus.Logger,
	appName string,
	defaultConfig map[string]interface{},
) (*koanf.Koanf, error) {
	f := pflag.NewFlagSet("config", pflag.ContinueOnError)
	f.Usage = func() {
		logger.Info(f.FlagUsages())

		os.Exit(0)
	}
	f.StringSlice(
		"conf",
		[]string{fmt.Sprintf("%s.toml", strings.ToLower(appName))},
		"path to one or more .toml files",
	)

	var err error

	if err = f.Parse(os.Args[1:]); err != nil {
		logger.WithError(err).Fatal("could not parse command line arguments")
	}

	config := koanf.New(".")

	_ = config.Load(confmap.Provider(defaultConfig, "."), nil)

	if err = config.Load(file.Provider(defaultFile), toml.Parser()); err != nil && !os.IsNotExist(err) {
		logrus.WithError(err).WithField("file", defaultFile).Fatal("error loading configuration from file")
	}

	cFiles, _ := f.GetStringSlice("conf")
	for _, c := range cFiles {
		if err = config.Load(file.Provider(c), toml.Parser()); err != nil {
			logger.WithError(err).WithField("file", c).Fatal("error loading configuration from file")
		}
	}

	if err = config.Load(posflag.Provider(f, ".", config), nil); err != nil {
		logger.WithError(err).Fatal("error loading configuration from command line")
	}

	prefix := fmt.Sprintf("%s_", strings.ToUpper(appName))

	if err = config.Load(env.Provider(prefix, ".", func(s string) string {
		return strings.ReplaceAll(strings.ToLower(strings.TrimPrefix(s, prefix)), "_", ".")
	}), nil); err != nil {
		logrus.WithError(err).Fatal("error loading configuration from environment")
	}

	return config, nil
}
